import greenfoot.Actor;
import java.awt.event.MouseEvent;
import java.awt.Point;
import java.awt.MouseInfo;
import java.util.List;

/**
 * This class works like a greenfoot.MouseInfo object. The coordinates are relative to the world size so you don't have to change the values you already got.
 * This class can be used the same way like greenfoot.MouseInfo objects without any changes.
 * 
 * @author Gevater_Tod4711
 * @version 1.0
 */
public class FullScreenMouseInfo
{
    private MouseEvent mouseEvent;
    private Point mousePoint;
    
    public FullScreenMouseInfo(MouseEvent mouseEvent) throws IllegalArgumentException {
        if (mouseEvent == null) {
            throw new IllegalArgumentException("The MouseEvent used for the FullScreenMouseInfo mussn't be null.");
        }
        this.mouseEvent = mouseEvent;
        mousePoint = mouseEvent.getLocationOnScreen();
    }
    
    /**
     * Return the actor (if any) that the current mouse behaviour is related to.
     * 
     * @return
     *      The actor this mouse behaviour is related to or null if there is no such actor.
     */
    public Actor getActor() {
        List<Actor> actors = FullScreenWindow.getDisplayedWorld().getObjectsAt(getX(), getY(), Actor.class);
        if (actors != null) {
            return actors.get(0);
        }
        return null;
    }
    
    /**
     * The number of the pressed or clicked button (if any).
     * 
     * @return
     *      The button number. Usually 1 is the left button, 2 is the middle button and 3 is the right button.
     */
    public int getButton() {
        if (mouseEvent != null) {
            return mouseEvent.getButton();
        }
        return -1;
    }
    
    /**
     * Return the number of mouse clicks associated with this mouse event.
     * 
     * @return
     *      The number of times a button has been clicked.
     */
    public int getClickCount() {
        if (mouseEvent != null) {
            return mouseEvent.getClickCount();
        }
        return -1;
    }
    
    /**
     * Return the current x position of the mouse cursor in the world.
     * The returned value is relative to the worlds size and not to the screen size so this method can be used like in a normal greenfoot scenario.
     * 
     * @return
     *      The x position in grid coordinates of the world.
     */
    public int getX() {
        return (int) (mousePoint.getX() * FullScreenWorld.WORLD_WIDTH / FullScreenWindow.FRAME_WIDTH);
    }
    
    /**
     * Return the current y position of the mouse cursor in the world.
     * The returned value is relative to the worlds size and not to the screen size so this method can be used like in a normal greenfoot scenario.
     * 
     * @return
     *      The y position in grid coordinates of the world.
     */
    public int getY() {
        return (int) (mousePoint.getY() * FullScreenWorld.WORLD_HEIGHT / FullScreenWindow.FRAME_HEIGHT);
    }
    
    /**
     * Return the current mouse point consisting of the x and y coordinates of the mouse.
     * The returned value is relative to the worlds size and not to the screen size so this method can be used like in a normal greenfoot scenario.
     * 
     * @return
     *      The current mouse point.
     */
    public Point getMousePoint() {
        return new Point(getX(), getY());
    }
    
    /**
     * The representing String for this FullScreenMouseInfo object.
     */
    public String toString() {
        return "FullScreenMouseInfo: (" + getX() + "|" + getY() + ") \t" + mouseEvent.toString();
    }
}
